#!/usr/bin/env python

"""
Utilities for working with coordinates of various types, notably ones in
hexagonal space.

Contains:

  * Constants/Functions for directions in the space that the chips/boards sit.
  * Constants for identifying edges of a hexagon of chips (uses same functions
    as directions)
  * Functions for addressing in a hexagonal world.
  * Functions for generating arrangements of hexagons.
  * Functions for Generation arrangements of spinnaker boards
  * Functions for transforming Cartesian coordinates
  * Functions for working with cabinets

This uses the hexagonal addressing scheme suggested in

	Addressing and Routing in Hexagonal Networks with Applications for Tracking
	Mobile Users and Connection Rerouting in Cellular Networks by Nocetti et. al.
"""

from enum import IntEnum

from spinner import coordinates

################################################################################
# Directions
################################################################################

class Direction(IntEnum):
	east       = 0
	north_east = 1
	north      = 2
	west       = 3
	south_west = 4
	south      = 5
	
	@property
	def next_ccw(self):
		"""
		Returns the next direction counter-clockwise from the given direction.
		"""
		return Direction((self + 1) % 6)
	
	@property
	def next_cw(self):
		"""
		Returns the next direction clockwise from the given direction.
		"""
		return Direction((self - 1) % 6)
	
	@property
	def opposite(self):
		"""
		Returns the opposite direction.
		"""
		return Direction((self + 3) % 6)
	
	@property
	def vector(self):
		"""
		Returns the vector which moves one unit in the given direction.
		"""
		return _DIRECTION_VECTORS[self]

_DIRECTION_VECTORS = {
	Direction.east:       ( 1, 0, 0),
	Direction.west:       (-1, 0, 0),
	Direction.north:      ( 0, 1, 0),
	Direction.south:      ( 0,-1, 0),
	Direction.north_east: ( 0, 0,-1),
	Direction.south_west: ( 0, 0, 1),
}


################################################################################
# Coordinates in hexagon world :)
################################################################################

def add_direction(vector, direction):
	"""
	Returns the vector moved one unit in the given direction.
	"""
	return coordinates.Hexagonal(*(v + a for (v,a) in zip(vector, direction.vector)))


def manhattan(vector):
	"""
	Calculate the Manhattan distance required to traverse the given vector.
	"""
	return sum(map(abs, vector))


def median_element(values):
	"""
	Returns the value of the median element of the set.
	"""
	return sorted(values)[len(values)//2]


def to_shortest_path(vector):
	"""
	Converts a vector into the shortest-path variation.
	
	A shortest path has at least one dimension equal to zero and the remaining two
	dimensions have opposite signs (or are zero).
	"""
	assert(len(vector) == 3)
	
	# The vector (1,1,1) has distance zero so this can be added or subtracted
	# freely without effect on the destination reached. As a result, simply
	# subtract the median value from all dimensions to yield the shortest path.
	median = median_element(vector)
	return coordinates.Hexagonal(*(v - median for v in vector))


def to_xy(vector):
	"""
	Takes a 3D vector and returns the equivalent 2D version.
	"""
	return coordinates.Hexagonal2D(vector[0] - vector[2], vector[1] - vector[2])


def hex_to_cartesian(coords):
	"""
	Convert a set of hexagonal coordinates into equivalent (for presentation
	purposes) Cartesian values.
	"""
	
	old_x, old_y = to_xy(coords)
	
	new_x = old_x
	new_y = (old_y * 2) - old_x
	
	return coordinates.Cartesian2D(new_x, new_y)


def board_to_chip(coords, layers = 4):
	"""
	Convert a hexagonal board coordinate into a hexagonal chip coordinate for the
	chip at the bottom-left of a board that size.
	"""
	x,y = hex_to_skewed_cartesian(coords)
	x *= layers
	y *= layers
	return coordinates.Hexagonal(x,y,0)


def hex_to_skewed_cartesian(coords):
	"""
	Convert a set of hexagonal coordinates into equivalent Cartesian values
	skewed to make x and y in the coordinate space match x and y in Cartesian
	space.
	"""
	
	old_x, old_y = to_xy(coords)
	
	new_x = old_x + old_y
	new_y = (old_y * 2) - old_x
	
	return coordinates.Cartesian2D(new_x, new_y)


def wrap_around(coord, bounds):
	"""
	Wrap the coordinate given around the edges of a torus made of hexagonal
	pieces. Assumes that the world is a NxM arrangement of threeboards (see
	threeboards function) with bounds = (N, M).
	
	XXX: Implementation could be nicer (and non iterative)...
	"""
	
	w,h = bounds
	x,y = to_xy(coord)
	
	assert(w > 0)
	assert(h > 0)
	
	while True:
		# Where is the coordinate relative to the world
		left  = x + y     < 0
		right = x + y     >= w*3
		below = (2*y) - x <  0
		above = (2*y) - x >= h*3
		
		if below and left:
			x += 1 + (w-1)*2 - (h-1)
			y += 2 + (w-1)   + (h-1)
			continue
		
		if above and right:
			x -= 1 + (w-1)*2 - (h-1)
			y -= 2 + (w-1)   + (h-1)
			continue
		
		if left:
			x += w*2
			y += w
			continue
		
		if right:
			x -= w*2
			y -= w
			continue
		
		if below:
			x -= h
			y += h
			continue
		
		if above:
			x += h
			y -= h
			continue
		
		break
	
	return coordinates.Hexagonal(x,y,0)




################################################################################
# Cartesian Coordinate Manipulations
################################################################################


def euclidean(v):
	"""
	The Euclidean distance represented by the given vector.
	"""
	return sum(x**2 for x in v) ** 0.5


def fold_dimension(x, w, f):
	r"""
	Takes a coordinate, x, on a single dimension of length w. Returns a tuple
	(new_x, fold) where new_x is the coordinate after the dimension has been
	folded into f pieces and fold is the fold number it is on.
	
	Input::
		
		 ______ w _____
		|              |
		-----+----------
		     |
		     x
		
		      | |
		     \| |/
		      \./
		
		new_x ___\__.\    f = 3
		          \/| \
		            |
		          fold
	"""
	# Width of the folded sections (round up so sections are bigger than
	# neccessary if not evenly divisible)
	fold_width = (w+(f-1)) // f
	
	new_x = x % fold_width
	fold  = x // fold_width
	
	# If on a reverse-facing fold, flip the coordinate
	if fold%2:
		# If this is the last fold, it may be smaller if not evenly divisible
		if fold == f - 1:
			new_x = (fold_width - ((fold_width*f) - w)) - new_x - 1
		else:
			new_x = fold_width - new_x - 1
	
	return (new_x, fold)


def fold_interleave_dimension(x, w, f):
	r"""
	As fold_dimension but returns a new x such that if the following points were
	folded, they would be mapped like so::
		
		 _______   ---\  (0,0) \  / (0,1)  ---\   _______
		 0 1 2 3   ---/   (1,0) \/ (1,1)   ---/   0 3 1 2
	
	That is, it interleaves points which would be mapped to the same position by
	fold_dimension.
	"""
	new_x, fold = fold_dimension(x,w,f)
	
	new_x *= f
	new_x += fold
	
	return new_x


################################################################################
# Cabinets
################################################################################


def cabinetise(coord, bounds, num_cabinets, frames_per_cabinet, boards_per_frame = None):
	r"""
	Takes a set of Cartesian coordinates and maps them into a series of cabinets.
	Splits the system into columns, one per cabinet. Splits each column into rows,
	one per frame. These rows likely consist of several columns and rows in
	Cartesian space and so values are interleaved to yield a board allocation.
	
	If the width of the set of boards doesn't divide into the num_cabinets
	or the height doesn't divide into the number of frames_per_cabinet, the axes
	are flipped and tried again. If this doesn't solve the problem, a ValueError
	is raised.
	
	coord is an (x,y) tuple containing the coordinate to map
	
	bounds is a (w,h) tuple containing the width and height of the Cartesian space
	
	If boards_per_frame is given then an assertion checks that the number of boards
	is adequate.
	
	Returns a tuple (cabinet, frame, board).
	"""
	
	x, y = coord
	w, h = bounds
	
	# If not divisible, try flipping the axes
	if w % num_cabinets != 0 or h % frames_per_cabinet != 0:
		y, x = x, y
		h, w = w, h
	
	# If still not divisible into cabinets and frames, fail.
	if w % num_cabinets != 0 or h % frames_per_cabinet != 0:
		raise ValueError("Cannot directly map boards into cabinets.")
	
	cols_per_cabinet = w // num_cabinets
	rows_per_frame   = h // frames_per_cabinet
	
	cabinet = x // cols_per_cabinet
	frame   = y // rows_per_frame
	
	# Sub coordinate within the frame
	x %= cols_per_cabinet
	y %= rows_per_frame
	
	# Interleave into board number
	board = y + (rows_per_frame * x)
	
	# Sanity check the position is actually within the system.
	assert(board < boards_per_frame)
	assert(frame < frames_per_cabinet)
	assert(cabinet < num_cabinets)
	
	return coordinates.Cabinet(cabinet, frame, board)



################################################################################
# Hexagon Generation
################################################################################

def hexagon(layers = 4):
	"""
	Generator which produces a list of (x,y) tuples which produce a hexagon of the
	given number of layers.
	
	Try me::
	
		points = set(hexagon(4))
		for y in range(min(y for (x,y) in points), max(y for (x,y) in points) + 1)[::-1]:
			for x in range(min(x for (x,y) in points), max(x for (x,y) in points) + 1):
				if (x,y) in points:
					print "#",
				else:
					print " ",
			print
	"""
	
	X,Y,Z = 0,1,2
	
	next_position = [0,0,0]
	
	for n in range(layers):
		for _ in range(n):
			yield to_xy(next_position)
			next_position[Y] -= 1
		
		for _ in range(n):
			yield to_xy(next_position)
			next_position[Z] += 1
		
		for _ in range(n+1):
			yield to_xy(next_position)
			next_position[X] -= 1
		
		for _ in range(n):
			yield to_xy(next_position)
			next_position[Y] += 1
		
		for _ in range(n+1):
			yield to_xy(next_position)
			next_position[Z] -= 1
		
		for _ in range(n+1):
			yield to_xy(next_position)
			next_position[X] += 1


def hexagon_zero(layers = 4):
	"""
	As with hexagon except coordinates are given relative to the bottom-left
	coordinate of the hexagon.
	"""
	
	return ( coordinates.Hexagonal2D(x+layers, y+layers-1)
	         for (x,y) in hexagon(layers)
	       )




################################################################################
# Threeboard Generation
################################################################################


def threeboards(width = 1, height = None):
	r"""
	Generates a list of width x height threeboards. If height is not specified,
	height = width. Width defaults to 1. Coordinates are given as (x,y,0) tuples
	on a hexagonal coordinate system like so::
		
		
		    | y
		    |
		   / \
		z /   \ x
	
	A threeboard looks like so::
		
		   ___
		  / 1 \___
		  \___/ 2 \
		  / 0 \___/
		  \___/
	
	With the bottom-left hexagon being at (0,0).
	
	And is tiled in to long rows like so::
		
		   ___     ___     ___     ___
		  / 0 \___/ 1 \___/ 2 \___/ 3 \___
		  \___/ 0 \___/ 1 \___/ 2 \___/ 3 \
		  / 0 \___/ 1 \___/ 2 \___/ 3 \___/
		  \___/   \___/   \___/   \___/
	
	And into a mesh like so::
		
		   ___     ___     ___     ___
		  / 4 \___/ 5 \___/ 6 \___/ 7 \___
		  \___/ 4 \___/ 5 \___/ 6 \___/ 7 \
		  / 4 \___/ 5 \___/ 6 \___/ 7 \___/
		  \___/ 0 \___/ 1 \___/ 2 \___/ 3 \___
		      \___/ 0 \___/ 1 \___/ 2 \___/ 3 \
		      / 0 \___/ 1 \___/ 2 \___/ 3 \___/
		      \___/   \___/   \___/   \___/
	
	"""
	
	height = width if height is None else height
	
	# Create the boards
	for y in range(height):
		for x in range(width):
			# z is the index of the board within the set. 0 is the bottom left, 1 is
			# the top, 2 is the right
			for z in range(3):
				# Offset within threeboard --+
				# Y offset ---------+        |
				# X offset -+       |        |
				#           |       |        |
				x_coord = (x*2) + (-y) + (z >= 2)
				y_coord = (x  ) + ( y) + (z >= 1)
				yield coordinates.Hexagonal(x_coord,y_coord,0)



from setuptools import setup, find_packages


# Defines __version__
with open("spinner/version.py", "r") as f:
	exec(f.read())


setup(
	name="spinnaker_spinner",
	version=__version__,
	packages=find_packages(),
	
	# Metadata for PyPi
	author="Jonathan Heathcote",
	description="A tool for generating wiring instructions for large SpiNNaker machines.",
	long_description = (
		"See `GitHub <https://github.com/SpiNNakerManchester/SpiNNer>`_ or "
		"the documentation on `ReadTheDocs <http://spinner.readthedocs.org/>`_ for "
		"more information."
	),
	license="GPLv2",
	
	# Requirements
	install_requires=[
		# For Python 2/3 cross-compatibility
		"six", "enum-compat",
		# For generating diagrams
		"cairocffi",
		# For displaying diagrams interactively (used to connect Cairo images to
		# tkinter)
		"Pillow",
		# For interacting with a SpiNNaker machine
		"rig",
	],
	
	# Scripts
	entry_points={
		"console_scripts": [
			"spinner-topology-stats = spinner.scripts.topology_stats:main",
			"spinner-wiring-stats = spinner.scripts.wiring_stats:main",
			"spinner-wiring-diagram = spinner.scripts.wiring_diagram:main",
			"spinner-wiring-guide = spinner.scripts.wiring_guide:main",
			"spinner-wiring-list = spinner.scripts.wiring_list:main",
			"spinner-wiring-validator = spinner.scripts.wiring_validator:main",
			"spinner-machine-map = spinner.scripts.machine_map:main",
			"spinner-ethernet-chips = spinner.scripts.ethernet_chips:main",
			"spinner-proxy-server = spinner.scripts.proxy_server:main",
		],
	}
)



"""Utilities for probing the wiring in a SpiNNaker machine."""

from six import iteritems

import random

from spinner.topology import Direction


class WiringProbeError(Exception):
	"""An error occurred when trying to probe the machine!"""
	pass


class WiringProbe(object):
	"""
	An object which probes a set of SpiNNaker boards to discover the connectivity
	between them. This tool uses the debugging facility of the board-to-board
	links to transmit/receive small values in their idle packets. As a result, the
	system should not be booted while performing link probing to ensure the links
	remain idle.
	"""
	
	# Number of bits in an ID which can be applied to a board
	NUM_ID_BITS = 16
	
	
	# A dictionary {direction: (fpga_num, reg_base_addr), ...} which gives for
	# each link direction the index of the responsible FPGA and the prefix for
	# register bank addresses for the particular link hardware within that FPGA.
	BOARD_FPGAS = {
		Direction.south:      (0, 0x00010000),
		Direction.east:       (0, 0x00000000),
		Direction.west:       (1, 0x00010000),
		Direction.south_west: (1, 0x00000000),
		Direction.north_east: (2, 0x00010000),
		Direction.north:      (2, 0x00000000),
	}
	
	# Register bank addresses for the "IDle Sentinel Input/Output" registers. The
	# IDSO register specifies a value to be transmitted as part of link idle
	# packets. IDSI indicates the value transmitted in the most recently received
	# idle packet.
	IDSO_REG = 19
	IDSI_REG = 20
	
	# Register address of the idle packet scrambling enable register
	SCRM_REG = 0x00040010
	
	# Register bank address of the handshake status bits. If bit 0 is set, the
	# link is alive, if 0, the link is dead.
	HAND_REG = 21
	
	def __init__(self, bmp_controller,
	             num_cabinets, frames_per_cabinet, boards_per_frame):
		"""Takes a connection to the system's BMPs along with the dimensions of the
		system."""
		self.bmp_controller = bmp_controller
		
		self.num_cabinets = num_cabinets
		self.frames_per_cabinet = frames_per_cabinet
		self.boards_per_frame = boards_per_frame
		
		# A dictionary {link_id: ((c,f,b),d), ...} and visa versa.
		self.id_to_link = {}
		self.link_to_id = {}
		
		# Assign every link an ID (also serves to check all FPGAs are powered on)
		self._assign_link_ids()
	
	def _assign_link_ids(self):
		"""Assign a unique ID to each HSS endpoint.
		
		Assigns IDs and loads them onto the machine. Raises a WiringProbeError if
		any FPGA is powered down.
		"""
		# Links are allocated sequential IDs and so values left by previous runs of
		# the tool could be the same or similar. To make this less likely, a fixed
		# random mask is XOR'd into every ID.
		mask = random.getrandbits(WiringProbe.NUM_ID_BITS)
		
		link_index = 0
		for c in range(self.num_cabinets):
			for f in range(self.frames_per_cabinet):
				for b in range(self.boards_per_frame):
					for d in Direction:
						# Generate the ID
						id = link_index ^ mask
						link_index += 1
						
						# Record it
						self.id_to_link[id] = (c,f,b,d)
						self.link_to_id[(c,f,b,d)] = id
						
						# Write it to the hardware
						self._write_register(c, f, b, d, WiringProbe.IDSO_REG, id)
						
						# Check the FPGA is powered on by reading back the register and
						# comparing it
						if self._read_register(c, f, b, d, WiringProbe.IDSO_REG) != id:
							raise WiringProbeError("FPGA not powered on "
							                       "(cabinet:{} frame:{} "
							                       "board:{} link:{})".format(c,f,b,d.name))
					
					# Also turn off scrabmling so this ID is actually sent out with idle
					# packets
					for fpga_num in range(3):
						self.bmp_controller.write_fpga_reg(fpga_num, WiringProbe.SCRM_REG, 0,
						                                   c, f, b)
	
	def _write_register(self, cabinet, frame, board, direction,
	                    reg_num, value):
		"""Write to a register in the specified link endpoint."""
		fpga_num, reg_base_addr = WiringProbe.BOARD_FPGAS[direction]
		addr = (reg_num<<2) | reg_base_addr
		
		self.bmp_controller.write_fpga_reg(fpga_num, addr, value,
		                                   cabinet, frame, board)
	
	
	def _read_register(self, cabinet, frame, board, direction, reg_num):
		"""Read a register in the specified link endpoint."""
		fpga_num, reg_base_addr = WiringProbe.BOARD_FPGAS[direction]
		addr = (reg_num<<2) | reg_base_addr
		
		return self.bmp_controller.read_fpga_reg(fpga_num, addr,
		                                         cabinet, frame, board)
	
	
	def get_link_target(self, cabinet, frame, board, direction):
		"""Determine which link is at the other end of the supplied link.
		
		Returns (cabinet, frame, board, direction) of the socket on the remote
		board. Returns None if the link is not connected to any known endpoint.
		"""
		# If the link is down, the link is definately not connected
		handshake = self._read_register(cabinet, frame, board, direction,
		                                WiringProbe.HAND_REG)
		if handshake & 1 == 0:
			return None
		else:
			id = self._read_register(cabinet, frame, board, direction,
			                         WiringProbe.IDSI_REG)
			return self.id_to_link.get(id, None)
	
	
	def discover_wires(self):
		"""
		Returns a list [(src, dst), ...] where src and dst are tuples (cabinet,
		frame, board, direction) specifying which links were found to be working (in both
		directions). Directions north, east and south_west are always "sources" and
		south, west, and north_east are always "destinations" except when an invalid
		connection (e.g. east to north) is made in which case the order is not
		defined.
		"""
		
		from_wires = set([])
		to_wires   = set([])
		
		for fc in range(self.num_cabinets):
			for ff in range(self.frames_per_cabinet):
				for fb in range(self.boards_per_frame):
					for fd in Direction:
						source = (fc, ff, fb, fd)
						target = self.get_link_target(fc, ff, fb, fd)
						if target is None:
							# Not connected...
							continue
						
						# Flip the from/to so that the from is going from north, east,
						# south_west (if neither end of the wire is from these then things are
						# very weird (connectors are electrically polarised so this isn't
						# possible) so just put up with the wrong order here).
						if fd in [Direction.south, Direction.west, Direction.north_east]:
							target, source = source, target
							wires = to_wires
						else:
							wires = from_wires
						
						wires.add((source, target))
		
		# Return only those wires which were discovered travelling in both
		# directions
		return list(from_wires & to_wires)



#!/usr/bin/env python

"""
Transformations on the coordinates to be applied to [(board, coord),...] lists.
"""

from operator import mul

from collections import defaultdict

from six import iteritems

from spinner import topology
from spinner import coordinates


def _assert_coord(boards, coordinate_types):
	"""
	Used Internally.
	
	Assert that the coordinates are for a Cartesian space (or that the set of
	boards is empty).
	"""
	assert(len(boards) == 0 or isinstance(boards[0][1], coordinate_types))


def hex_to_cartesian(boards):
	"""
	Convert hexagonal coordinates into 2D Cartesian coordinates maintaining the
	shape of the input. That is, feed in a mesh from board.create_torus() which
	looks like a rhombus and the resulting set of Cartesian coordinates will also
	look like a torus.
	"""
	_assert_coord(boards, coordinates.Hexagonal)
	
	return [ (board, topology.hex_to_cartesian(coord))
	         for (board, coord) in boards
	       ]


def hex_to_skewed_cartesian(boards):
	"""
	Convert hexagonal coordinates into 2D Cartesian coordinates skewing the input
	-30deg on the x-axis. That is, feed in a mesh from board.create_torus() which
	looks like a rhombus and the resulting set of Cartesian coordinates will look
	like a ragged rectangle.
	"""
	_assert_coord(boards, coordinates.Hexagonal)
	
	return [ (board, topology.hex_to_skewed_cartesian(coord))
	         for (board, coord) in boards
	       ]


def rhombus_to_rect(boards):
	r"""
	Performs a modulo max+1 for all coordinates. When given, for e.g., the rhombus
	arrangement of a toroid of 3-boards (e.g. from create_torus()) which has been
	mapped onto Cartesian coordinates turns it into a rectangle like so::
		
		_________         ___   ______             _________
		\        \        \  | |      \           |         |
		 \        \   -->  \ | |       \      --> |         |
		  \._______\        \| |._______\ .       |.________|
		   |                             /;\       |
		 (0,0)               \___________/       (0,0)
	"""
	_assert_coord(boards, (coordinates.Cartesian2D, coordinates.Cartesian3D))
	
	# Can't do anything with an empty input
	if len(boards) == 0:
		return []
	
	maxes = tuple(map(max, *(c for (b,c) in boards)))
	return [ (board, type(boards[0][1])(*(v%(m+1) for (v,m) in zip(c, maxes))))
	         for (board, c) in boards
	       ]


def compress(boards, x_div = 1, y_div = 2):
	r"""
	Compress coordinates, more precisely, does integer division on coordinates.
	This is useful for taking the hexagonal pattern in Cartesian coordinates and
	making it a more regular pattern like so::
		
		     ___     ___
		 ___/ 9 \___/11 \               +---+---+---+---+
		/ 8 \___/10 \___/               | 8 | 9 |10 |11 |
		\___/ 5 \___/ 7 \    ------\    +---+---+---+---+
		/ 4 \___/ 6 \___/    ------/    | 4 | 5 | 6 | 7 |
		\___/ 1 \___/ 3 \               +---+---+---+---+
		/ 0 \___/ 2 \___/               | 0 | 1 | 2 | 3 |
		\___/   \___/                   +---+---+---+---+
	
	Here, compress(x=1, y=2) takes all coordinates and maps them to (x/1, y/2)
	using integer division.
	"""
	_assert_coord(boards, coordinates.Cartesian2D)
	
	return [ (board, coordinates.Cartesian2D(int(x)//int(x_div), int(y)//int(y_div)))
	         for (board, (x,y)) in boards
	       ]


def flip_axes(boards):
	"""
	Invert the axes of the array of boards.
	
	Useful when attempting to cabinetise a grid which doesn't divide unless the
	axes are flipped.
	"""
	_assert_coord(boards, coordinates.Cartesian2D)
	
	return [ (board, coordinates.Cartesian2D(y, x))
	         for (board, (x, y)) in boards
	       ]


def fold(boards, folds):
	r"""
	Takes a set of Cartesian coordinates and folds into the number of segments
	specified for each dimension in folds. The folded segments are then
	interleaved. Below, fold(b, (2,1)) is shown::
		
		+---+---+---+---+          +---+---+---+---+
		| 8 | 9 |10 |11 |          | 8 |11 | 9 |10 |
		+---+---+---+---+  -----\  +---+---+---+---+
		| 4 | 5 | 6 | 7 |  -----/  | 4 | 7 | 5 | 6 |
		+---+---+---+---+          +---+---+---+---+
		| 0 | 1 | 2 | 3 |          | 0 | 3 | 1 | 2 |
		+---+---+---+---+          +---+---+---+---+
	"""
	_assert_coord(boards, (coordinates.Cartesian2D, coordinates.Cartesian3D))
	
	# Can't do anything with an empty input
	if len(boards) == 0:
		return []
	
	# Must have a number of folds and gaps for each dimension
	assert(len(boards[0][1]) == len(folds))
	
	maxes = tuple(map(max, *(c for (b,c) in boards)))
	
	# Use topology.fold_dimension() to get the fold number and multiply this by
	# the gap size to get an offset for each value.
	return [ (board, type(boards[0][1])(*[topology.fold_interleave_dimension(v,m+1,f)
	                                      for (v,m,f)
	                                      in zip(c, maxes, folds)]
	                                   ))
	         for (board, c) in boards
	       ]


def cabinetise(boards, num_cabinets, frames_per_cabinet, boards_per_frame = None):
	r"""
	Takes a set of Cartesian coordinates and maps them into a series of cabinets.
	Splits the system into columns, one per cabinet. Splits each column into rows,
	one per frame. These rows likely consist of several columns and rows in
	Cartesian space and so values are interleaved to yield a board allocation.
	
	If the width of the set of boards doesn't divide into the num_cabinets
	or the height doesn't divide into the number of frames_per_cabinet, the axes
	are flipped and tried again. If this doesn't solve the problem, a ValueError
	is raised.
	
	If boards_per_frame is given then an assertion checks that the number of boards
	is adequate.
	"""
	_assert_coord(boards, coordinates.Cartesian2D)
	
	# Can't do anything with an empty input
	if len(boards) == 0:
		return []
	
	max_x, max_y = map(max, *(c for (b,c) in boards))
	
	return [ (board, topology.cabinetise( (x,y)
	                                    , (max_x+1, max_y+1)
	                                    , num_cabinets
	                                    , frames_per_cabinet
	                                    , boards_per_frame
	                                    ))
	         for (board, (x,y)) in boards
	       ]


def remove_gaps(boards):
	"""
	Take a cabinetised system and shift boards right in their frames to remove any
	empty gaps between boards in frames.
	"""
	_assert_coord(boards, coordinates.Cabinet)
	
	# Initially collect together boards which share a frame
	# {(c, f): {b: Board, ...}, ...}
	frames = defaultdict(dict)
	for board, (c, f, b) in boards:
		frames[(c, f)][b] = board
	
	# Reconstruct the mapping one frame at a time, placing all baords into
	# contiguous blocks.
	boards = []
	for (c, f), rack_boards in iteritems(frames):
		# Renumber the boards in the frame
		b = 0
		for old_b, board in sorted(iteritems(rack_boards),
			                         key=(lambda tup: tup[0])):
			boards.append((board, coordinates.Cabinet(c, f, b)))
			b += 1
	return boards


def cabinet_to_physical(boards, cabinet):
	"""
	Takes Cabinet coordinates and converts them into Cartesian3D coordinates
	representing the physical positions of cabinets based on a
	spinner.cabinet.Cabinet specification.
	"""
	_assert_coord(boards, coordinates.Cabinet)
	
	return [(board, cabinet.get_position(*coord)) for (board,coord) in boards]



#!/usr/bin/env python

"""Generate illustrations of SpiNNaker machine wiring."""

import argparse

from spinner.diagrams.machine import MachineDiagram

from spinner.topology import Direction

from spinner.utils import folded_torus

from spinner import transforms
from spinner import topology

from spinner.scripts import arguments

from spinner.scripts.contexts import PDFContextManager, PNGContextManager


def add_diagram_arguments(parser):
	"""
	Add arguments which specify the output format and what is to be drawn.
	"""
	parser.add_argument("--wire-thickness", choices="thick normal thin".split(),
	                    default="normal",
	                    help="set the thickness of wires drawn (default: "
	                         "%(default)s)")
	
	parser.add_argument("--highlight", action=arguments.CabinetAction(4, append=True),
	                    help="highlight a particular cabinet/frame/board/socket "
	                         "with a red border")
	
	parser.add_argument("--hide-labels", "-L", action="store_true", default=False,
	                    help="hide socket/board/frame/cabinet number labels")


def get_diagram_arguments(parser, args, w, h, cabinet, num_frames):
	"""
	Validate and unpack the arguments added by add_diagram_arguments.
	"""
	# Determine the set of cabinets/frames/boards to focus on and the aspect ratio
	# of the resulting part of the system (to enable automatic image width/height
	# determination later)
	if cabinet.num_cabinets == 1 and num_frames == 1:
		focus = (0, 0, slice(0, 3 * w * h))
		dimensions = cabinet.get_dimensions(boards=3 * w * h)
	elif cabinet.num_cabinets == 1:
		focus = (0, slice(0, num_frames))
		dimensions = cabinet.get_dimensions(frames=num_frames)
	else:
		focus = (slice(0, cabinet.num_cabinets), )
		dimensions = cabinet.get_dimensions()
	aspect_ratio = float(dimensions.x) / float(dimensions.y)
	
	wire_thickness = args.wire_thickness
	highlights = args.highlight or []
	hide_labels = args.hide_labels
	
	return (aspect_ratio, focus, wire_thickness, highlights, hide_labels)


def main(args=None):
	parser = argparse.ArgumentParser(
		description="Generate illustrations of SpiNNaker machine wiring.")
	arguments.add_version_args(parser)
	arguments.add_image_args(parser)
	add_diagram_arguments(parser)
	arguments.add_topology_args(parser)
	arguments.add_cabinet_args(parser)
	arguments.add_subset_args(parser)
	
	# Process command-line arguments
	args = parser.parse_args(args)
	(w, h), transformation, uncrinkle_direction, folds =\
		arguments.get_topology_from_args(parser, args)
	
	cabinet, num_frames =\
		arguments.get_cabinets_from_args(parser, args)
	
	aspect_ratio, focus, wire_thickness, highlights, hide_labels =\
		get_diagram_arguments(parser, args, w, h, cabinet, num_frames)
	
	output_filename, file_type, image_width, image_height =\
		arguments.get_image_from_args(parser, args, aspect_ratio)
	
	wire_filter = arguments.get_subset_from_args(parser, args)
	
	# Generate folded system
	hex_boards, folded_boards = folded_torus(w, h,
	                                         transformation,
	                                         uncrinkle_direction,
	                                         folds)
	
	# Divide into cabinets
	cabinetised_boards = transforms.cabinetise(folded_boards,
	                                           cabinet.num_cabinets,
	                                           num_frames,
	                                           cabinet.boards_per_frame)
	cabinetised_boards = transforms.remove_gaps(cabinetised_boards)
	
	# Set up diagram
	md = MachineDiagram(cabinet)
	
	# Create lookup from cabinet coord to chip x/y to enable labelling of boards
	b2cab = dict(cabinetised_boards)
	b2chip = dict((b, topology.to_xy(topology.board_to_chip(c)))
	              for b, c in hex_boards)
	cab2chip = dict((b2cab[b], b2chip[b]) for b, c in cabinetised_boards)
	
	# Add labels
	if not hide_labels:
		for cabinet_num in range(cabinet.num_cabinets):
			md.add_label(cabinet_num, cabinet_num)
			for frame_num in range(cabinet.frames_per_cabinet):
				md.add_label(frame_num, cabinet_num, frame_num)
				for board_num in range(cabinet.boards_per_frame):
					# Only label boards which are actually part of the system
					xy = cab2chip.get((cabinet_num, frame_num, board_num), None)
					if xy is not None:
						md.add_label("{} ({},{})".format(board_num, xy.x, xy.y),
						             cabinet_num, frame_num, board_num)
						for socket in Direction:
							name = "".join(w[0] for w in socket.name.split("_")).upper()
							md.add_label(name, cabinet_num, frame_num, board_num, socket,
							             rgba=(1.0, 1.0, 1.0, 0.7))
	
	# Add highlights
	for highlight in highlights:
		md.add_highlight(*highlight, width=cabinet.board_dimensions.x/3.0)
	
	# Add wires
	wire_thickness_m = {
		"thick" : cabinet.board_dimensions.x / 5.0,
		"normal" : cabinet.board_dimensions.x / 10.0,
		"thin" : cabinet.board_dimensions.x / 20.0,
	}[wire_thickness]
	b2c = dict(cabinetised_boards)
	for direction in [Direction.north, Direction.west, Direction.north_east]:
		for b, c in cabinetised_boards:
			ob = b.follow_wire(direction)
			oc = b2c[ob]
			
			src = (c.cabinet, c.frame, c.board, direction)
			dst = (oc.cabinet, oc.frame, oc.board, direction.opposite)
			if wire_filter((src, dst)):
				md.add_wire(src, dst, width=wire_thickness_m)
	
	# Render the image
	Context = {"png": PNGContextManager,
	           "pdf": PDFContextManager}[file_type]
	with Context(output_filename, image_width, image_height) as ctx:
		md.draw(ctx, image_width, image_height, *(((list(focus) + [None]*3)[:3])*2))
	
	return 0


if __name__=="__main__":  # pragma: no cover
	import sys
	sys.exit(main())




r"""A proxy enabling multiple wiring guide instances to interact with the same
SpiNNaker boards.

A very simple protocol is used between the client and server. Clients may send
the following new-line delimited commands to the server:

* ``VERSION,[versionstring]\n`` The server will disconnect any client with an
  incompatible version number reported for ``[versionstring]``. Returns
  ``OK\n``.
* ``LED,[c],[f],[b],[lednum],[state]\n`` Turn on or off the specified LED. Note
  that the LED remains switched on while *any* client wants it to be on.
  Returns ``OK\n``.
* ``TARGET,[c],[f],[b],[link]\n`` Discover what link is at the other end of the
  supplied link. Returns ``[c],[f],[b],[link]\n`` or ``None\n`` if no link is
	connected. Note that links are represented by their number, not their name.
"""

import traceback

import socket

import select

from collections import defaultdict

import logging

from six import iteritems

from spinner.version import __version__

from spinner.topology import Direction


DEFAULT_PORT = 6512


class ProxyError(Exception):
	"""Exception raised when the proxy cannot connect."""
	pass


class ProxyServer(object):
	"""A proxy server enabling multiple wiring guide instances to interact with
	the same SpiNNaker boards.
	"""
	
	def __init__(self, bmp_controller, wiring_probe,
	             hostname="", port=DEFAULT_PORT):
		self.bmp_controller = bmp_controller
		self.wiring_probe = wiring_probe
		
		# Open a TCP socket
		self.server_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self.server_sock.setsockopt(socket.SOL_SOCKET,
                                socket.SO_REUSEADDR, 1)
		self.server_sock.bind((hostname, port))
		self.server_sock.listen(5)
		
		self.client_socks = []
		
		# A buffer for unprocessed data received from each client
		self.client_buffer = {}
		
		# For each LED, maintains a set of clients which have turned it on
		self.led_setters = defaultdict(set)
	
	
	def add_client(self, sock, addr):
		"""Register a new client."""
		logging.info("New connection {} from {}".format(sock, addr))
		self.client_socks.append(sock)
		
		# Create buffer for received data (and schedule its deletion upon
		# disconnection)
		self.client_buffer[sock] = b""
	
	
	def remove_client(self, sock):
		"""Disconnect and cleanup after a particular child."""
		logging.info("Closing socket {}".format(sock))
		
		# Remove buffer
		self.client_buffer.pop(sock)
		
		# Turn off any LEDs left on by the client
		for (c, f, b, led), socks in iteritems(self.led_setters):
			if sock in socks:
				self.set_led(sock, c, f, b, led, False)
		
		# Close socket
		self.client_socks.remove(sock)
		sock.close()
	
	
	def set_led(self, sock, c, f, b, led, state):
		"""Set the state of a diagnostic LED.
		
		An LED is turned on if at least one client has turned it on. An LED is only
		turned off if all clients which have turned the LED on have also turned it
		off again.
		"""
		
		setters = self.led_setters[(c, f, b, led)]
		
		cur_led_state = bool(setters)
		
		if state:
			setters.add(sock)
		else:
			setters.discard(sock)
		
		new_led_state = bool(setters)
		
		if cur_led_state != new_led_state:
			self.bmp_controller.set_led(led, new_led_state, c, f, b)
	
	
	def handle_version(self, sock, args):
		"""Handle "VERSION" commands.
		
		This command contains, as the argument, the SpiNNer version number of the
		remote client. If the version of the client does not match the server, the
		client is disconnected.
		
		Arguments: vX.Y.Z
		
		Returns: OK
		"""
		# Check for identical version
		assert args.decode("ascii") == __version__
		sock.send(b"OK\n")
	
	
	def handle_led(self, sock, args):
		"""Handle "LED" commands.
		
		Set the state of a diagnostic LED on a board.
		
		Arguments: c,f,b,led,state
		
		Returns: OK
		"""
		c, f, b, led, state = map(int, args.split(b","))
		
		self.set_led(sock, c, f, b, led, state)
		
		sock.send(b"OK\n")
	
	
	def handle_target(self, sock, args):
		"""Handle "TARGET" commands.
		
		Determine what is at the other end of a given link.
		
		Arguments: c,f,b,d
		
		Returns: c,f,b,d or None
		"""
		c, f, b, d = map(int, args.split(b","))
		
		target = self.wiring_probe.get_link_target(c, f, b, d)
		
		if target is None:
			sock.send(b"None\n")
		else:
			sock.send("{},{},{},{}\n".format(*map(int, target)).encode("ascii"))
	
	
	def process_data(self, sock, data):
		"""Process data received from a socket."""
		# Prepend any previously unprocessed data
		data = self.client_buffer[sock] + data
		
		# Handle any received commands. If a command fails (or is invalid) the
		# connection is dropped.
		try:
			while b"\n" in data:
				line, _, data = data.partition(b"\n")
				logging.debug("Handling command {} from {}".format(line, sock))
				
				cmd, _, args = line.partition(b",")
				
				# If an unrecognised command arrives, this lookup will fail and get
				# caught by the exception handler, printing an error and disconnecting
				# the client.
				{
					b"VERSION": self.handle_version,
					b"LED": self.handle_led,
					b"TARGET": self.handle_target,
				}[cmd](sock, args)
		
		except Exception as e:
			logging.exception(
				"Disconnected client {} due to bad command (above)".format(sock))
			self.remove_client(sock)
			return
		
		# Retain any remaining unprocessed data
		self.client_buffer[sock] = data
	
	
	def main(self):
		logging.info("Starting proxy server...")
		
		try:
			while True:
				ready, _1, _2 = select.select([self.server_sock] + self.client_socks, [], [])
				
				for sock in ready:
					if sock is self.server_sock:
						# New client connected!
						self.add_client(*self.server_sock.accept())
					else:
						# Data arrived from a client
						try:
							data = sock.recv(1024)
						except (IOError, OSError) as exc:
							logging.error(
								"Socket {} failed to receive: {}".format(sock, exc))
							# Cause socket to get closed
							data = b""  # pragma: no branch
							
						if len(data) == 0:
							# Connection closed
							self.remove_client(sock)
						else:
							self.process_data(sock, data)
		except KeyboardInterrupt:
			# Disconnect all clients (also cleans up LED states, etc.)
			for sock in self.client_socks:
				self.remove_client(sock)
		
		logging.info("Proxy server terminated cleanly.")


class ProxyClient(object):
	"""A client for the ProxyServer object defined above.
	
	This object implements a BMPController-compatible ``set_led`` method and
	WiringProbe compatible ``get_link_target`` method and thus may be substituted
	for the above when these functions are all that are required, e.g. for the
	InteractiveWiringGuide.
	"""
	
	def __init__(self, hostname, port=DEFAULT_PORT):
		"""Connect to a running ProxyServer."""
		self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self.sock.connect((hostname, port))
		
		# A receive buffer
		self.buf = b""
		
		# Check for protocol version compatibility.
		self.check_version()
	
	def recvline(self):
		"""Wait for a full line to be received from the server."""
		while b"\n" not in self.buf:
			data = self.sock.recv(1024)
			self.buf += data
			if len(data) == 0:
				raise ProxyError("Remote server closed the connection.")
		
		line, _, self.buf = self.buf.partition(b"\n")
		return line
	
	def check_version(self):
		"""Check that the remote server has a compatible protocol version."""
		self.sock.send("VERSION,{}\n".format(__version__).encode("ascii"))
		
		if self.recvline() != b"OK":
			raise ProxyError("Remote server has incompatible protocol version")
	
	def set_led(self, led, state, c, f, b):
		"""Set the state of an LED on the remote machine."""
		self.sock.send("LED,{},{},{},{},{}\n".format(
			c, f, b, led, int(state)).encode("ascii"))
		
		if self.recvline() != b"OK":
			raise ProxyError("Got unexpected response to LED command.")
	
	def get_link_target(self, c, f, b, d):
		"""Discover the other end of a specified link on a remote machine."""
		self.sock.send("TARGET,{},{},{},{}\n".format(
			c, f, b, int(d)).encode("ascii"))
		
		response = self.recvline()
		if response == b"None":
			return None
		else:
			try:
				c, f, b, d = map(int, response.split(b","))
				return (c, f, b, Direction(d))
			except ValueError:
				raise ProxyError("Got unexpected response to TARGET command.")



"""Utility functions for generating wiring plans."""

from math import ceil, sqrt

from spinner import board
from spinner import topology
from spinner import coordinates
from spinner import transforms


def ideal_system_size(num_boards):
	"""Calculate the ideal system size for a system with the specified number of
	boards.
	
	Returns
	-------
	(w, h)
		Width and height in triads of the most square system possible with the given
		number of boards. When a square system cannot be made, the function prefers
		wider systems over taller systems on the basis that most cabinet
		arrangements are much wider than they are tall.
	
	Raises
	------
	TypeError
		If the number of boards is not a multiple of three. (Hexagonal toruses
		require multiples of three boards to construct).
	"""
	if num_boards % 3 != 0:
		raise TypeError("{} is not a multiple of 3".format(num_boards))
	
	# Special case to avoid division by 0
	if num_boards == 0:
		return (0, 0)
	
	# Find the largest pair of factors to discover the squarest system
	for h in reversed(range(1, int(sqrt(num_boards//3)) + 1)):  # pragma: no branch
		if (num_boards//3) % h == 0:
			break
	
	w = (num_boards//3) // h
	
	return (w, h)


def folded_torus(w, h, transformation, uncrinkle_direction, folds):
	"""Generate a 2D arrangement of boards in a (w, h) triad torus folded in the
	specified fashion.
	
	Parameters
	----------
	w : int
		Width of the system in triads
	h : int
		Width of the system in triads
	transformation : "slice" or "shear"
		The transformation to use to map from hexagonal coordinates to a rectangular
		2D grid.
	uncrinkle_direction : "rows" or "columns"
		When uncrinckling the hexagonal mesh into a 2D grid, should groups rows be
		flattened or groups of columns. Typically, rows will be chosen.
	folds : (x, y)
		Number of peieces to fold each dimension into.
	
	Returns
	-------
	(hex_boards, folded_boards)
		hex_boards is a list of tuples (board, Hexagonal(x, y)) giving the logical
		coordinates for each board on a 2D hexagonal grid.
		
		folded_boards is a list of tuples (board, Cartesian2D(x, y)) giving the coordinates of
		the boards laid out such that wirelength is minimised.
	"""
	# Boards in their Hexagonal coordinate system
	hex_boards = board.create_torus(w, h)
	
	if uncrinkle_direction not in ("rows", "columns"):
		raise TypeError("Uncrinkle direction must be 'rows' or 'columns'")
	rows = uncrinkle_direction == "rows"
	cols = uncrinkle_direction == "columns"
	
	if transformation == "slice":
		cart_boards = transforms.rhombus_to_rect(
			transforms.compress(
				transforms.hex_to_cartesian(
					hex_boards),
				x_div=2 if cols else 1, y_div=2 if rows else 1))
	elif transformation == "shear":
		cart_boards = transforms.compress(
				transforms.hex_to_skewed_cartesian(
					hex_boards),
				x_div=3 if cols else 1, y_div=3 if rows else 1)
	else:
		raise TypeError("Unsupported transformation {}".format(transformation))
	
	# Fold and interleave
	folded_boards = transforms.fold(cart_boards, folds)
	
	return (hex_boards, folded_boards)


def min_num_cabinets(num_boards, frames_per_cabinet, boards_per_frame):
	"""Calculate the minimum number of cabinets and frames required to house the
	given set of boards.
	
	Just simple arithmetic!
	
	Returns
	-------
	(num_cabinets, num_frames)
	"""
	
	num_cabinets = int(ceil(num_boards / float(frames_per_cabinet *
	                                           boards_per_frame)))
	# If more than one cabinet is required, we must be using all the frames in
	# each cabinet
	if num_cabinets > 1:
		return (num_cabinets, frames_per_cabinet)
	
	num_frames = int(ceil(num_boards / float(boards_per_frame)))
	
	return (num_cabinets, num_frames)




"""Markdown generation utilities for producing reports."""

def heading(text, level=1):
	"""Render the supplied text as a heading.
	
	Parameters
	----------
	text : str
		The text to render as a heading
	level : int
		The heading level from 1 upward.
	"""
	if level == 1:
		return "{}\n{}\n".format(text, "="*len(text))
	elif level == 2:
		return "{}\n{}\n".format(text, "-"*len(text))
	else:
		return "{} {}\n".format("#"*level, text)


def table(data):
	"""Render the supplied data as a table.
	
	Parameters
	----------
	data : [[object, ...], ...]
		Table data, a list of rows of values. Each row must be the same length.
		Values are formated with `str`. The first row of data is presented as a
		header.
	"""
	# Special case if no data provided
	if len(data) == 0:
		return "\n"
	
	# Convert all data into strings
	data = [[str(v) for v in row] for row in data]
	
	# Work out the maximum width of each column
	col_widths = [0]*len(data[0])
	for row in data:
		for n, col in enumerate(row):
			if len(col) > col_widths[n]:
				col_widths[n] = len(col)
	
	# Print the table
	out = ""
	for row_num, row in enumerate(data):
		out += "| {} |\n".format(
			" | ".join(
				col.ljust(col_widths[col_num])
				for col_num, col in enumerate(row)
			)
		)
		
		# Print underline for header row
		if row_num == 0:
			out += "| {} |\n".format(
				" | ".join("-"*w for w in col_widths)
			)
	
	return out
